# This class subclasses the string built-in and allows for interpolation of values.
from __future__ import nested_scopes
import exceptions
#from spark import GenericScanner, Token

__all__ = ['istring', 'TRUE', 'FALSE', 'SELF', 'CALLER']

TRUE = 1
FALSE = 0
# These constants represent the place in the stack from which to pull variables, from the
# perspective of the caller.
SELF = 2
CALLER = 3

class IStringError(exceptions.Exception): pass
class InterpolatedVariableNotDefined(IStringError):
    def __init__(self, interp_key):
        self._interp_key = interp_key
    def __str__(self):
        result = "Attempted to use undefined variable: %s\n" % (self._interp_key)
        return result
class InvalidSplitError(IStringError):
    def __init__(self, num, orig, pat, result=None):
        self._num = num
        self._orig = orig
        self._pat = pat
        self._result = result
    def __str__(self):
        return ("An invalid number of results (%s) were obtained splitting string `%s' " \
            + "using pattern `%s'") % (self._num, self._orig, self._pat)
class IllegalFunctionCallError(IStringError):
    def __init__(self, params):
        self._params = params
    def __str__(self):
        return ("Function or method call uses parameters; parameter are not allowed.  Parameters " \
            + "used = `%s'") % self._params
class EvalFailedError( IStringError ):
    def __init__( self, value ):
        self._value = value
    def __str__( self ):
        return "The variable or function name `%s' could not be evaluated." % self._value

#class IStringScanner(GenericScanner):

#    def __init__(self):
#        GenericScanner.__init__(self)

#    def tokenize(self, input):
#        self.rv = []
#        GenericScanner.tokenize(self, input)
#        return self.rv

#    def t_whitespace(self, s):
#        r' \s+'
#        pass

#    def t_op(self, s):
#        r'\$ | \&'
#        self.rv.append(Token(type=s))

class istring(str):

    def __new__(cls, value='', depth=2):
        from types import StringType, UnicodeType
        variables = _extract_names(depth)
        istr, vars = substitute(value, variables)
        if type(istr) == StringType:
            self = str.__new__(cls, istr)
        elif type(istr) == UnicodeType:
            self = unicode.__new__(uistring, istr)
        else:
            raise TypeError
        self.vars = vars
        return self

    def __init__(self, value='', depth=2):
        self.raw_str = value

    def __getattr__(self, name):
        if name == '_raw_str':
            return str()
        else:
            raise AttributeError

    def _reload(self):
        """Return the current interpolated string."""
        return istring(self.raw_str, depth=3)

    # Define accessors and properties
    def set_raw_str(self, value):
        """Set the raw (pre-interpolated) string value."""
        self._raw_str = value

    def get_raw_str(self):
        """Get the raw (pre-interpolated) string value."""
        return self._raw_str

    def set_vars(self, vars):
        self._vars = vars

    def get_vars(self):
        return self._vars

    # Class properties
    raw_str = property(get_raw_str, set_raw_str)
    istr = property(_reload, None)
    vars = property(get_vars, set_vars)

# I have to maintain two separate classes for string and unicode strings because I can't
# multiply inherit from base classes (i.e., str and unicode).  You shouldn't instantiate
# something directly from uistring.  Just pass a unicode string into istring.
class uistring(unicode):

    def __new__(cls, value='', depth=2):
        from types import StringType, UnicodeType
        variables = _extract_names(depth)
        istr, vars = substitute(value, variables)
        if type(istr) == StringType:
            self = str.__new__(istring, istr)
        elif type(istr) == UnicodeType:
            self = unicode.__new__(cls, istr)
        else:
            raise TypeError
        self.vars = vars
        return self

    def __init__(self, value='', depth=2):
        self.raw_str = value

    def __getattr__(self, name):
        if name == '_raw_str':
            return str()
        else:
            raise AttributeError

    def _reload(self):
        """Return the current interpolated string."""
        return istring(self.raw_str, depth=3)

    # Define accessors and properties
    def set_raw_str(self, value):
        """Set the raw (pre-interpolated) string value."""
        self._raw_str = value

    def get_raw_str(self):
        """Get the raw (pre-interpolated) string value."""
        return self._raw_str

    def set_vars(self, vars):
        self._vars = vars

    def get_vars(self):
        return self._vars

    # Class properties
    raw_str = property(get_raw_str, set_raw_str)
    istr = property(_reload, None)
    vars = property(get_vars, set_vars)

def eval_word(variables, control_chars, word):
    from string import split
    import re
    import sys
    import NameMapper
    import neo.istring.istring_std as nis
    MATCH_PATS = [r'(\$\$\.)$', r'(\$\.)$', r'(\.)$', r'(\$\$)$', r'(\$)$', r'(\&\&)$',
                  r'(\&)$', r'()$']
    IS_CONTEXT_PATS = ['$$.', '$$', '.', '&&']
    INTERPOLATE_PATS = ['$.', '$', '$$.', '$$']
    STRING_PATS = ['$.', '$$.']
    WORD_PREFIX = r'(.*?)(\\*)'
    FN_SEGMENTS = r'(.*?)\((.*)\)'
    INTEGER = r'\d+'
    FLOAT = r'\d*\.\d+'
    STRING = r'[\"\'](.*)[\"\']'
    ENDING_PARENS = r'^(.*)\((.*)\)$'
    for pat in MATCH_PATS:
        m = re.search(WORD_PREFIX + pat, control_chars)
        if m:
            # first_used_char holds the literal character that may have been escaped.  If
            # it was escaped, it should not be considered an operator (such as $ or $$),
            # but rather a literal part of the string.
            first_used_char = str()
            num_esc_was_even = FALSE
            unused_chars = m.group(1)
            escapes = m.group(2)
            used_chars = m.group(3)
            # Check word for ending parens.  Strip them if it's just a simple open-and-close
            # paren, otherwise raise an error.
            paren_groups = re.search(ENDING_PARENS, word)
            if paren_groups and used_chars not in ['&', '&&']:
                if paren_groups.group(2):
                    raise IllegalFunctionCallError(paren_groups.group(2))
                else:
                    word = paren_groups.group(1)
            if len(escapes) % 2:
                if used_chars:
                    first_used_char = used_chars[0]
                    used_chars = used_chars[1:]
                escapes = escapes[1:]
            else:
                num_esc_was_even = TRUE
            # Determine whether the end result should be quoted.
            quote_result = 0
            while unused_chars[-2:] == '$.' and used_chars in IS_CONTEXT_PATS:
                if unused_chars[-3:] == '$$.':
                    quote_result += 1
                    unused_chars = unused_chars[:-3]
                else:
                    quote_result += 1
                    unused_chars = unused_chars[:-2]
            # %s will be replaced with escapes later
            prefix = unused_chars + '%s' + first_used_char
            try:
                if used_chars in INTERPOLATE_PATS:
                    word_val = NameMapper.valueForName(variables, word, executeCallables=TRUE)
                    if used_chars in STRING_PATS:
                        word_val = "'%s'" % word_val
                elif used_chars in ['&', '&&']:
                    # Get parameter values and execute the function.
                    params = []
                    fn_segs = re.search(FN_SEGMENTS, word)
                    fn = fn_segs.group(1)
                    args = split(fn_segs.group(2), ',')
                    for a in args:
                        try:
                            params.append(NameMapper.valueForName(variables, a, executeCallables=TRUE))
                        except NameMapper.NotFound:
                            int_match = re.search(INTEGER, a)
                            float_match = re.search(FLOAT, a)
                            str_match = re.search(STRING, a)
                            if int_match:
                                params.append(int(a))
                            elif float_match:
                                params.append(float(a))
                            elif str_match:
                                params.append(str_match.group(1))
                            else:
                                params.append(a)
                    word_val = apply(nis.__dict__[fn], tuple(params))
                else:  # FIXME handle &, &&
                    word_val = word
            except NameMapper.NotFound:
                sys.stderr.write("INTERPOLATION ERROR OCCURRED.  String:\n")
                sys.stderr.write("%s\n" % word)
                raise InterpolatedVariableNotDefined(word)
            except KeyError:
                word_val = word
            if not used_chars:
                if escapes:
                    is_context = TRUE
                else:
                    is_context = FALSE
            elif used_chars in IS_CONTEXT_PATS:
                if not len(unused_chars):
                    is_context = TRUE
                else:
                    if unused_chars == '.' and not len(escapes):
                        is_context = TRUE
                    elif used_chars == '&' and not len(escapes):
                        is_context = TRUE
                    else:
                        is_context = FALSE
                if used_chars == '.':
                    prefix += used_chars
            else:
                is_context = FALSE
            if is_context and num_esc_was_even:
                escapes = escapes[:len(escapes)/2]
            prefix = prefix % escapes
            if quote_result:
                quotes = quote_result * "'"
                return "%s%s%s" % (quotes, prefix + str(word_val), quotes), is_context
            else:
                return prefix + str(word_val), is_context
    else:
        return control_chars + word, FALSE

def get_words(segments, eval_self_only, context_str):
    """Fetch the next three sections from segments, or default values."""
     # Note that segments.pop() is guaranteed not to have any \, $, & or . chars.
    if eval_self_only:
        word = segments.pop()
    else:
        word = segments.pop() + context_str
    try:
        control_chars = segments.pop()
    except IndexError:
        control_chars = str()
    try:
        prev_word = segments[-1]
    except IndexError:
        prev_word = str()
    return word, control_chars, prev_word

def handle_parens(interp_word, remainder):
    """Make sure that if interp_word contains parentheses, that these are either of canonical
    form, i.e. function(.*?), or truncate any text with parentheses in it."""
    import re
    from string import index
    HAS_PAREN_PAIR = r'^(.*?)\((.*?)\)(.*)'
    m = re.search(HAS_PAREN_PAIR, interp_word)
    if m:
        remainder = m.group(3) + remainder
        interp_word = "%s(%s)" % (m.group(1), m.group(2))
    else:
        try:
            left_paren_pos = index(interp_word, '(')
        except ValueError:
            left_paren_pos = len(interp_word)
        try:
            right_paren_pos = index(interp_word, ')')
        except ValueError:
            right_paren_pos = len(interp_word)
        if left_paren_pos < right_paren_pos:
            earliest_pos = left_paren_pos
        else:
            earliest_pos = right_paren_pos
        remainder = interp_word[earliest_pos:] + remainder
        interp_word = interp_word[:earliest_pos]
    return interp_word, remainder

def handle_quotes(interp_word, remainder):
    from string import index
    try:
        single_quote_pos = index(interp_word, '\'')
    except ValueError:
        single_quote_pos = len(interp_word)
    try:
        double_quote_pos = index(interp_word, '\"')
    except ValueError:
        double_quote_pos = len(interp_word)
    if single_quote_pos < double_quote_pos:
        str_pos = single_quote_pos
    else:
        str_pos = double_quote_pos
    remainder = interp_word[str_pos:] + remainder
    interp_word = interp_word[:str_pos]
    return interp_word, remainder

def substitute(s, variables=None):
    """Substitute interpolated variables in a string using the passed dictionary.  Assume
    passive interpolation unless escaping ($$) is used.  E.g.:

    >>> m[34] = 'thirty-four'
    >>> m[3] = 'three'
    >>> a = 4
    >>> istring("$m.3$$a")
    "thirty-four"

    To interpolate the above string, we first evaluate $a which yeilds 4.  Then we
    interpolate m.34 which yeilds "thirty-four".  By contrast:

    >>> istring("$m.3$a")
    "three4"

    The double dollar sign indicates that $a should be interpreted as a separate variable.
    A double dollar sign is only necessary when a variable is not preceeded by whitespace
    (i.e., when the immediately preceeding character does not match [\w_\.]).  String-based
    escaping uses the $$. sequence:

    >>> istring($a$$.m.3")
    "4'three'"
    """
    from neo.utilities import set_cmp
    from string import join
    import re

    if not variables:
        variables = _extract_names()
    #print "keys in variables: %s" % join(variables.keys(), ',')

    WORD_PAT = r'([\\\$\&\.]+)'
    FIRST_ALPHA_SEG = r'^([\w_\.\(\)\'\"]+)(.*)'
    ALL_CHARS_ALPHA = r'^[\w_]+$'
    replacements = dict()
    processed_str = str()
    context_str = str()
    segments = re.split(WORD_PAT, s)
    # Handle trivial cases.  Two elements should be impossible.  One element means
    # our pattern didn't show up, so there's nothing to interpolate.
    if len(segments) == 2 or not len(segments):
        raise InvalidSplitError(len(segments), s, WORD_PAT)
    if len(segments) == 1: return segments[0], replacements
    eval_self_only = TRUE
    while len(segments) >= 1:
        word, control_chars, prev_word = get_words(segments, eval_self_only, context_str)
        # If the current word does not begin with alphanumeric characters, then no attempt
        # is being made to interpolate.  E.g.: "$ foo"; $ would be the control chars and
        # " foo" would be the current word.
        m = re.search(FIRST_ALPHA_SEG, word, re.S)
        if not m:
            processed_str = control_chars + word + processed_str
            eval_self_only = TRUE
            continue
        interp_word = m.group(1)
        remainder = m.group(2)
        # Make sure the interp_word is properly formatted if parens are involved.
        cmp_val = set_cmp(interp_word, '()', '\'\"')
        # The remainder may be reset below.  These functions should return the complete,
        # correct remainder....this is not an accidental overwriting of the remainder
        # variable.
        if cmp_val == 1:
            interp_word, remainder = handle_parens(interp_word, remainder)
        elif cmp_val == -1:
            interp_word, remainder = handle_quotes(interp_word, remainder)
        if control_chars:
            old_word = interp_word
            try:
                interp_word, is_context = eval_word(variables, control_chars, interp_word)
            except AttributeError:
                raise EvalFailedError( interp_word )
            replacements[old_word] = interp_word
        else:
            is_context = FALSE
        # Interpret this word in context with the previous words only if all the control
        # characters are interpreted, AND if they aren't '$$' or '$$.', AND if the previous
        # word contains only alphanumeric characters.
        # Note that if m evaluates true, there is a previous word (i.e. len(segments) >= 1).
        m = re.search(ALL_CHARS_ALPHA, prev_word, re.S)
        if m and is_context:
            if eval_self_only:
                context_str = interp_word + remainder + context_str
            else:
                context_str = interp_word + remainder
            eval_self_only = FALSE
        else:
            eval_self_only = TRUE
            context_str = str()
            processed_str = interp_word + remainder + processed_str
    return processed_str, replacements

def _extract_names(levels=2):
    import sys
    l_globals = sys._getframe(levels).f_globals
    l_locals  = sys._getframe(levels).f_locals
    variables = {}
    variables.update(l_globals)
    variables.update(l_locals)
    return variables

def regression_tests(indentation=0):
    from neo.common_testing import test_eq, test_ne, test_exc, disp
    from neo.istring.istring_std import regression_tests as i_std_regression
    import re

    def functest(): return 'func'
    class classtest:
        def methtest(self): return 'method'

    disp("Running istring test suite...\n", indentation)
    i_std_regression(1)
    ############################################################################
    ### SLICES                                                               ###
    ############################################################################
    disp("    Testing slices... ", indentation)
    foo = istring("the quick brown fox")
    test_eq(foo[4:9], "quick", "failed!")
    ############################################################################
    ### REGULAR EXPRESSION USAGE                                             ###
    ############################################################################
    disp("    Testing usage in regular expressions... ", indentation)
    m = re.search('(quick)', foo)
    test_ne(m, None, "regex search failed!", last=FALSE)
    test_eq(m.group(1), "quick", "failed to fetch group!")
    ############################################################################
    ### INTEGER INTERPOLATION                                                ###
    ############################################################################
    disp("    Testing integer interpolation... ", indentation)
    j = 3
    test_eq(istring("int=$j"), "int=3", "failed!")
    ############################################################################
    ### FLOAT INTERPOLATION                                                  ###
    ############################################################################
    disp("    Testing float interpolation... ", indentation)
    k = 3.45
    test_eq(istring("float=$k"), "float=3.45", "failed!")
    ############################################################################
    ### STRING INTERPOLATION                                                 ###
    ############################################################################
    disp("    Testing string interpolation... ", indentation)
    l = "foo"
    l2 = "bar"
    test_eq(istring("str=$l"), "str=foo", "failed direct interpolation!", last=FALSE)
    test_eq(istring("str=$.l"), "str='foo'", "failed string-style interpolation!", last=FALSE)
    test_eq(istring("str$l2$.l$.l2"), "strbar'foo''bar'", "failed sequential string-style interpolation!")
    ############################################################################
    ### LIST INTERPOLATION                                                   ###
    ############################################################################
    disp("    Testing list interpolation... ", indentation)
    m = [1,2,3]
    test_eq(istring("list=$m.1"), "list=2", "failed literal interpolation!", last=FALSE)
    m = [1,2,3]
    m2 = 1
    test_eq(istring("list=$m.$$m2"), "list=2", "failed variable interpolation!")
    ############################################################################
    ### DICTIONARY INTERPOLATION                                             ###
    ############################################################################
    disp("    Testing dictionary interpolation... ", indentation)
    n = {1:9,2:8,3:7}
    test_eq(istring("dict=$n.2"), "dict=8", "failed literal interpolation!", last=FALSE)
    n = {1:9,2:8,3:7,'foo':'bar','space key':'yak'}
    n2 = 2
    test_eq(istring("dict=$n.$$n2"), "dict=8", "failed variable interpolation!", last=FALSE)
    oo = "oo"
    test_eq(istring("dict=$n.f$$oo"), "dict=bar", "failed mixed variable interpolation!", last=FALSE)
    test_eq(istring("dict=$n.foo"), "dict=bar", "failed direct interpolation!")
    #test_eq(istring("dict=$n.'space key'"), "dict=yak", "failed literal-value interpolation!")
    ############################################################################
    ### FUNCTION INTERPOLATION                                               ###
    ############################################################################
    disp("    Testing function interpolation... ", indentation)
    test_eq(istring("func=$functest"), "func=func", "failed!")
    ############################################################################
    ### CLASS METHOD INTERPOLATION                                           ###
    ############################################################################
    disp("    Testing class method interpolation... ", indentation)
    c = classtest()
    test_eq(istring("method=$c.methtest"), "method=method", "failed!", last=FALSE)
    test_eq(istring("method=$c.methtest()"), "method=method", "failed!")
    #test_exc("istring(\"method=$c.methtest()\")", IllegalFunctionCallError, "failed!")
    ############################################################################
    ### MULTIPLE INTERPOLATION                                               ###
    ############################################################################
    disp("    Testing multiple interpolations... ", indentation)
    test_eq(istring("int=$j;float=$k;str=$l;func=$functest;method=$c.methtest"), \
                "int=3;float=3.45;str=foo;func=func;method=method", "failed!")
    ############################################################################
    ### SEQUENTIAL INTERPOLATION                                             ###
    ############################################################################
    disp("    Testing sequential interpolations... ", indentation)
    test_eq(istring("$j$k$l$functest$c.methtest$m.2$n.3"), "33.45foofuncmethod37", "failed!")
    ############################################################################
    ### UNUSUAL CHARACTER SEQUENCES                                          ###
    ############################################################################
    disp("    Testing unusual character sequences... ", indentation)
    j = 3
    test_eq(istring("int=$^%@*(j$j"), "int=$^%@*(j3",
                "incorrect handling of miscellaneous characters!", last=FALSE)
    test_eq(istring("int='$j'"), "int='3'", "incorrect handling of single quote!", last=FALSE)
    test_eq(istring("int=\"$j\""), "int=\"3\"", "incorrect handling of double quote!", last=FALSE)
    test_eq(istring("int=)$j("), "int=)3(", "incorrect handling of close paren!", last=FALSE)
    test_eq(istring("int=($j)"), "int=(3)", "incorrect handling of open paren!", last=FALSE)
    a = "foo"
    b = "bar"
    test_eq(istring("This is 'a $a test\$' and a $b test"), "This is 'a foo test$' and a bar test",
                "failed period-quote sequence!", last=FALSE)
    test_eq(istring("This is 'a $a test\&' and a $b test"), "This is 'a foo test&' and a bar test",
                "failed period-quote sequence!", last=FALSE)
    test_eq(istring("This is 'a $a test$\$' and a $b test"), "This is 'a foo test$$' and a bar test",
                "failed period-quote sequence!", last=FALSE)
    test_eq(istring("This is 'a $a test.' and a $b test"), "This is 'a foo test.' and a bar test",
                "failed period-quote sequence!")
    ############################################################################
    ### ESCAPING                                                             ###
    ############################################################################
    disp("    Testing escaping... ", indentation)
    esc = "TeSt"
    val = "value"
    test_eq(istring(r"noesc=\$esc$$val"), "noesc=$escvalue", "failed $ escaping!", last=FALSE)
    test_eq(istring(r"noesc=$\$esc$$val"), "noesc=$$escvalue", "failed $$ escaping!", last=FALSE)
    test_eq(istring(r"noesc=\&upper(esc)$val"), "noesc=&upper(esc)value", "failed & escaping!", last=FALSE)
    test_eq(istring(r"noesc=$val\esc$$esc"), "noesc=valueescTeSt", "failed blend escaping!", last=FALSE)
    test_eq(istring(r"noesc=$val\\esc$$esc"), r"noesc=value\escTeSt", "failed double-escape escaping!")
    ############################################################################
    ### ATTRIBUTES                                                           ###
    ############################################################################
    disp("    Testing attributes... ", indentation)
    value = "value"
    test_eq(istring("value=$value").raw_str, "value=$value", "failed raw_str test!", last=FALSE)
    s = istring("value=$value")
    value = "newvalue"
    test_eq(s.istr, "value=newvalue", "failed istr test!", last=FALSE)
    test_eq(istring("value=$value").vars['value'], "newvalue", "failed vals test!")
    ############################################################################
    ### AMPERSAND INTERPOLATION                                              ###
    ############################################################################
    disp("    Testing ampersand interpolation... ", indentation)
    esc = "TeSt"
    test_eq(istring("value=&upper(esc)"), "value=TEST", "failed!", last=FALSE)
    test_eq(istring("value=&upper('TeSt')"), "value=TEST", "failed!", last=FALSE)
    test_eq(istring("value=$.&&upper(esc)"), "value='TEST'",
                "failed ampersand string interpolation!", last=FALSE)
    test_eq(istring("value=$.$.&&upper(esc)"), "value=''TEST''",
                "failed double ampersand string interpolation!", last=FALSE)
    test_eq(istring("value=$.&upper(esc)"), "value=$.TEST",
                "failed ampersand escaped interpolation!")
    ############################################################################
    ### STANDARD FUNCTIONS                                                   ###
    ############################################################################
    disp("    Testing standard functions... ", indentation)
    l = "a b c d".split()
    a = [1,2,3]
    b = [4,5,6]
    esc = "TeSt"
    test_eq(istring("upper=&upper(esc)"), "upper=TEST", "failed upper std function test", last=FALSE)
    test_eq(istring("lower=&lower(esc)"), "lower=test", "failed lower std function test", last=FALSE)
    test_eq(istring("len=&length(l)"), "len=4", "failed length std function test", last=FALSE)
    test_eq(istring("value=&list_to_str(l)"), "value=a,b,c,d", "failed list_to_str std function test", last=FALSE)
    test_eq(istring("value1=&list_to_str(a) value2=&list_to_str(b)"), "value1=1,2,3 value2=4,5,6",
                "failed multiple list_to_str calls!")
    ############################################################################
    ### COMBINATION WITH STRING FORMATTING                                   ###
    ############################################################################
    disp("    Testing in combination with string formatting... ", indentation)
    m = "abc"
    n = "def"
    test_dict = {'foo': 'bar', 'yak': 'lmnop'}
    test_eq(istring("value=%(foo)s$m,%(yak)s$n" % test_dict), "value=barabc,lmnopdef",
                "failed string formatting combination (1)!", last=FALSE)
    test_eq(istring("value=%(foo)s$m,%(yak)s$n") % test_dict, "value=barabc,lmnopdef",
                "failed string formatting combination (2)!")
    ############################################################################
    ### ERROR HANDLING                                                       ###
    ############################################################################
    disp( "    Testing error handling... ", indentation )
    try:
        print istring( "This should raise an exception &qrzvml" )
    except EvalFailedError:
        print "passed!"
    else:
        print "failed! (EvalFailedError not raised for invalid function)"
    ############################################################################
    ### UNICODE HANDLING                                                     ###
    ############################################################################
    #disp("    Testing unicode handling... ", indentation)
    #test_eq(type(istring(u"This is a test")), type(u"This is a test"), "failed to create unicode string!")

    disp("istring testing completed!\n", indentation)


if __name__ == '__main__':
    regression_tests()
